searchData={"items":[{"type":"module","title":"Datacop","doc":"An authorization library with `Dataloader` and `Absinthe` support.","ref":"Datacop.html"},{"type":"function","title":"Datacop.default_loader/1","doc":"Returns initialized dataloader struct with the source.\n\nIt requires `data/0` function to be defined for the particular module.","ref":"Datacop.html#default_loader/1"},{"type":"function","title":"Datacop.permit/4","doc":"Authorize an action.\n\nProcesses `c:Datacop.Policy.authorize/3` result.","ref":"Datacop.html#permit/4"},{"type":"function","title":"Examples - Datacop.permit/4","doc":"> Datacop.permit(MyApp.Accounts, :view_email, current_user, subject: other_user)\n    :ok\n\n    > Datacop.permit(MyApp.Accounts, :view_email, current_user, subject: other_user, loader: loader)\n    {:error, %Datacop.UnauthorizedError{message: \"Unauthorized\"}}","ref":"Datacop.html#permit/4-examples"},{"type":"function","title":"Datacop.permit?/4","doc":"The same as `permit/4`, but returns a `boolean`.","ref":"Datacop.html#permit?/4"},{"type":"type","title":"Datacop.action/0","doc":"Any term which describes an action you want to perform. Usually atoms are used.","ref":"Datacop.html#t:action/0"},{"type":"type","title":"Datacop.actor/0","doc":"Usually your user struct","ref":"Datacop.html#t:actor/0"},{"type":"type","title":"Datacop.option/0","doc":"Option\n\n* `:subject` – any value you want to access in the authorize/3 callback.\n* `:loader` – an initialized Dataloader struct with loaded sources.","ref":"Datacop.html#t:option/0"},{"type":"module","title":"Datacop.AbsintheMiddleware.Authorize","doc":"Performs authorization for the given resolution.\n\nThis module helps to build authorization logic for batches, based on [`Dataloader`](https://hexdocs.pm/dataloader/Dataloader.html).\nEach time the middleware receives `{:dataloader, _dataloader_config}` tuple, it loads(accumulates) data in the `dataloader`\nstruct, in order to run single query in the end. As a result, this middleware returns resolution with either\na successfully resolved state or Datacop.UnauthorizedError.","ref":"Datacop.AbsintheMiddleware.Authorize.html"},{"type":"module","title":"Options - Datacop.AbsintheMiddleware.Authorize","doc":"* `:loader` - either `dataloader` struct or 1-arity function, which can fetch `dataloader` struct,\n  based on `c:Absinthe.Schema.context/1`. It uses default `Datacop.default_loader/1`, when the option is nil.\n* `:actor` - the actor struct, which is used in the authorize/3 function for the target module.\n* `:subject` - the subject struct, which is used in the authorize/3 function for the target module.\n  By default this is resolution.source.\n* `:callback` - custom function callback, which handles Datacop result.","ref":"Datacop.AbsintheMiddleware.Authorize.html#module-options"},{"type":"module","title":"Example - Datacop.AbsintheMiddleware.Authorize","doc":"```elixir\nopts = [\n  loader: &(&1.loader),\n  actor: &(&1.actor),\n  callback: fn\n    :ok -> {:ok, true)\n    error -> {:ok, false}\n  end}\n]\nmiddleware(Authorize, {MyApp.Blog, :view_stats, opts})\n```","ref":"Datacop.AbsintheMiddleware.Authorize.html#module-example"},{"type":"type","title":"Datacop.AbsintheMiddleware.Authorize.opts/0","doc":"","ref":"Datacop.AbsintheMiddleware.Authorize.html#t:opts/0"},{"type":"behaviour","title":"Datacop.Policy","doc":"Where authorization rules live.\n\nTypically the callbacks are designed to be used by `Datacop.permit/4` and\nare not called directly.\n\nThe only requirement is to implement the `authorize/3` callback:\n\n    defmodule MyApp.MyContext do\n      @behaviour Datacop.Policy\n\n      def authorize(action, user, params) do\n        # Return :ok or true to permit\n        # Return {:error, reason}, or false to deny\n        # Return {:dataloader, data} to process data in the Dataloader\n      end\n    end\n\nTo perform authorization checks, use `Datacop.permit/4`:\n\n    with :ok <- Datacop.permit(MyApp.MyContext, :action_name, user, subject: :value) do\n      # ...\n    end\n\n    if Datacop.permit?(MyApp.MyContext, :action_name, user, subject: :value) do\n      # ...\n    end\n\nIf you want to define the callbacks in another module, you can use\n`defdelegate`:\n\n    defmodule MyApp.MyContext do\n      defdelegate authorize(action, user, params), to: MyApp.MyContext.Policy\n    end","ref":"Datacop.Policy.html"},{"type":"callback","title":"Datacop.Policy.authorize/3","doc":"","ref":"Datacop.Policy.html#c:authorize/3"},{"type":"type","title":"Datacop.Policy.dataloader_result/0","doc":"","ref":"Datacop.Policy.html#t:dataloader_result/0"},{"type":"exception","title":"Datacop.UnauthorizedError","doc":"","ref":"Datacop.UnauthorizedError.html"},{"type":"type","title":"Datacop.UnauthorizedError.t/0","doc":"","ref":"Datacop.UnauthorizedError.html#t:t/0"},{"type":"extras","title":"README","doc":"# Datacop\n[![Hex.pm](https://img.shields.io/hexpm/v/datacop.svg)](https://hex.pm/packages/datacop)\n\nAn authorization library with Dataloader and Absinthe support.\n\nThis library is heavily inspired by [bodyguard](https://hex.pm/packages/bodyguard).\nOur authorization rules not always simple, so `datacop` allows you to deal with n+1 queries using `dataloader`.","ref":"readme.html"},{"type":"extras","title":"Installation - README","doc":"The package can be installed by adding `datacop` and optionally `absinthe` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:datacop, \"~> 0.1\"},\n    {:absinthe, \"~> 1.6\"}\n  ]\nend\n```\nDocumentation can be found at [https://hexdocs.pm/datacop](https://hexdocs.pm/datacop).","ref":"readme.html#installation"},{"type":"extras","title":"Usage - README","doc":"#","ref":"readme.html#usage"},{"type":"extras","title":"Define a Policy module with authorization rules - README","doc":"This module should contain authorization rules or redirect resolution to `dataloader` for batch resolutions.\nTry to keep authorize callback pure and redirect side effects to `dataloader`.\n\n```elixir\ndefmodule MyApp.Blog.Policy do\n  @behaviour Datacop.Policy\n\n  @imp true\n  def authorize(:delete_post, actor, _post), do: actor.id == post.author_id or actor.admin?\n\n  def authorize(:view_stats, actor, post) do\n    if actor.admin? do\n      {:dataloader,\n       %{\n         source_name: MyApp.Blog,\n         batch_key: {:one, MyApp.Blog.Post},\n         inputs: [{{:can_admin_view_stats?, actor.id}, post.id}]\n       }}\n    else\n      false\n    end\n  end\nend\n```\n\n#","ref":"readme.html#define-a-policy-module-with-authorization-rules"},{"type":"extras","title":"Define a Data module for integration with `dataloader` - README","doc":"A typical module for working with `dataloader`. In this example we use [`Dataloader.Ecto`](https://hexdocs.pm/dataloader/Dataloader.Ecto.html).\nSee documentation for this module for detailed explanation how it works.\nBatch query should return a list of boolean values in the same order which `post_ids` has.\n```elixir\ndefmodule MyApp.Blog.Data do\n  def data do\n    Dataloader.Ecto.new(MyApp.Repo, run_batch: &run_batch/5)\n  end\n\n  def run_batch(queryable, _query, {:can_admin_view_stats?, admin_id}, post_ids, repo_opts, _params) do\n    result =\n      queryable\n      |> very_complex_query_returns_posts_which_are_managed_by_admin(admin_id, post_ids)\n      |> select([posts], {posts.id, true})\n      |> MyApp.Repo.all(repo_opts)\n      |> Map.new()\n\n    Enum.map(post_ids, &Map.get(result, &1, false))\n  end\nend\n```\n\n#","ref":"readme.html#define-a-data-module-for-integration-with-dataloader"},{"type":"extras","title":"Use context module as a proxy - README","doc":"It is not necessary to do this, but otherwise you'll have to refer to Data and Policy modules directly in places\nwhere corresponding functions are invoked.\n\n```elixir\ndefmodule MyApp.Blog do\n  defdelegate authorize(action, actor, params), to: __MODULE__.Policy\n  defdelegate data, to: __MODULE__.Data\nend\n```\n\n#","ref":"readme.html#use-context-module-as-a-proxy"},{"type":"extras","title":"Setup dataloader in Absinthe schema - README","doc":"See [this guide](https://hexdocs.pm/absinthe/dataloader.html#usage) for reference.\nIn general implementation of `c:Absinthe.Schema.context/1` should look like this:\n```elixir\ndef context(ctx) do\n  loader =\n    Dataloader.new() |> Dataloader.add_source(MyApp.Blog, MyApp.Blog.data())\n\n  Map.put(ctx, :loader, loader)\nend\n```\n\n#","ref":"readme.html#setup-dataloader-in-absinthe-schema"},{"type":"extras","title":"Use as a single action - README","doc":"Because absinthe defines `:loader` in `c:Absinthe.Schema.context/1` callback, we can reuse it in resolver functions by passing `:loader` option explicitly:\n```elixir\ndef delete_post(params, %{context: %{actor: actor, loader: loader}}) do\n  with {:ok, post} <- MyApp.Blog.fetch_post(params.post_id),\n       :ok <- Datacop.permit(MyApp.Blog, :delete_post, actor, subject: post, loader: loader) do\n    MyApp.Blog.delete_post(post, params)\n  end\nend\n```\nIf you don't pass `:loader`, then `datacop` checks if passed module (in example above it is `MyApp.Blog`) has `data/0` function. If yes, then loader can be lazily initiated by `datacop` for single source with passed module as a `:source_name`.\n\nFor our example this call will work:\n```elixir\nDatacop.permit(MyApp.Blog, :view_stats, actor, subject: post)\n```\nwhich is a short version of:\n```elixir\nDatacop.permit(MyApp.Blog, :view_stats, actor,\n  subject: post,\n  loader: Dataloader.new() |> Dataloader.add_source(MyApp.Blog, MyApp.Blog.data())\n)\n```\nbut this won't (`MyApp.Blog.Policy` doesn't implement `data/0`):\n```elixir\nDatacop.permit(MyApp.Blog.Policy, :view_stats, actor, subject: post)\n```\nThe next example works fine, as `:delete_post` action doesn't use dataloader:\n```elixir\nDatacop.permit(MyApp.Blog.Policy, :delete_post, actor, subject: post)\n```\nWith `Datacop.permit?/4` it's also possible to work with booleans:\n```elixir\nDatacop.permit?(MyApp.Blog, :search, actor, loader: loader)}\n```\n\n#","ref":"readme.html#use-as-a-single-action"},{"type":"extras","title":"Use as Absinthe middleware - README","doc":"In order to leverage full potential of `datacop` it is recommended to use it with `absinthe`.\n```elixir\nalias Datacop.AbsintheMiddleware.Authorize\n\nobject :post do\n  field :id, :id\n  field :stats, :stats do\n    middleware(Authorize, {MyApp.Blog, :view_stats, loader: &(&1.loader), actor: &(&1.actor)})\n    resolve(...)\n  end\nend\n```\nIn order to DRY you may want to provide a custom middleware on top of existing one:\n```elixir\ndefmodule MyApp.Schema.Middleware.Authorize do\n  @behaviour Absinthe.Middleware\n\n  @impl Absinthe.Middleware\n  def call(resolution, {action, context_module}) do\n    call(resolution, {action, context_module, []})\n  end\n\n  @impl Absinthe.Middleware\n  def call(resolution, {action, context_module, opts}) do\n    opts =\n      opts\n      |> Keyword.put_new(:actor, &(&1.actor))\n      |> Keyword.put_new(:loader, &(&1.loader))\n\n    params = {action, context_module, opts}\n\n    %{resolution | middleware: [{Authorization.AbsintheMiddleware.Authorize, params} | resolution.middleware]}\n  end\nend\n```\n\nand a helper on top of it\n```elixir\ndef authorize(action, module, opts \\\\ []) do\n  {:middleware, PtWeb.Schema.Middleware.Authorize, {action, module, opts}}\nend\n```\nso block with `:stats` contains less noise:\n```elixir\nfield :stats, :stats do\n  authorize(MyApp.Blog, :view_stats)\n  resolve(...)\nend\n```\n\nThat's it. Now if you request list of posts, then authorization will be performed in batches.","ref":"readme.html#use-as-absinthe-middleware"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,49,46,50]}}